%{

/*
 Parser for PL/LOLCODE
 $Id: gram.y,v 1.31 2008/04/30 10:45:41 eggyknap Exp $
*/

/* TODO: Make error messages return line number */
#define YYPARSE_PARAM result
#include "pllolcode.h"
#include "executor/spi.h"
#include <stdio.h>
#include <string.h>

int  pllolcode_yylex(void);
void pllolcode_yyerror(char *s);
int  pllolcode_yyparse(void *result);

void executeProgram(lolNodeList *prg);
bool executeList(lolNodeList *list);
bool executeNode(lolNode *node);
/*nodeStruct makeNode(NodeType typ, nodeStruct child1, nodeStruct child2, NodeVal nodeVal);
void printNodeTree(nodeStruct prg); */
lolNode *lolMakeNode(NodeType typ, NodeVal nodeVal, lolNodeList *list);
void printNodeList(lolNodeList *list);
void printNode(lolNode *cur);

void yarnNode(lolNode *node);
void numbrNode(lolNode *node);
void numbarNode(lolNode *node);
void troofNode(lolNode *node);
void visibleNode(lolNode *node);
void foundyrNode(lolNode *node);
void identNode(lolNode *node);
void declareNode(lolNode *node);
lolIdent arithNode(lolNode *node, lolIdent a, lolIdent b);
lolIdent boolNode(lolNode *node, lolIdent a, lolIdent b);
lolIdent comparisonNode(lolNode *node, lolIdent a, lolIdent b);
void binaryOprNode(lolNode *node);
void assignNode(lolNode *node);
void unaryOprNode(lolNode *node);
void allOfNode(lolNode *node);
void anyOfNode(lolNode *node);
void castNode(lolNode *node);
bool orlyNode(lolNode *node);
void wtfNode(lolNode *node);
void iminyrNode(lolNode *node);
void SPInode(lolNode *node);

NodeVal tmpnval;
lolNodeList *curList;
bool debugList = 0;
%}

/* BISON Declarations */ %name-prefix="pllolcode_yy"

%expect 29

%union  {
	int numbrVal;
	double numbarVal;
	char *yarnVal;
	struct lolNodeList *nodeList;
	struct lolNode *node;
}

%type <nodeList> lol_program lol_cmdblock lol_orly_block lol_wtf_block
%type <node> lol_const lol_var lol_cmd lol_xaryoprgroup lol_expression
%type <yarnVal> YARN
%type <yarnVal> IDENTIFIER
%type <numbrVal> NUMBR
%type <numbarVal> NUMBAR
%type <numbrVal> VISDEBUG VISLOG VISNOTICE VISINFO VISWARNING VISEXCEPTION lol_visibleLevel
%type <numbrVal> lol_binaryopr lol_unaryopr lol_xaryopr lol_typename

%token HAI KTHXBYE EXPREND
%token YARN IDENTIFIER NUMBR NUMBAR
%token TROOFWIN TROOFFAIL TROOF
%token VISIBLE VISDEBUG VISLOG VISNOTICE VISINFO VISWARNING VISEXCEPTION
%token FOUNDYR IHASA ITZ R
%token SUMOF DIFFOF PRODUKTOF QUOSHUNTOF MODOF BIGGROF SMALLROF AN
%token BOTHOF EITHEROF WONOF NOT ALLOF ANYOF MKAY
%token BOTHSAEM DIFFRINT
%token MAEK A TROOFTYPE NUMBRTYPE YARNTYPE NUMBARTYPE NOOBTYPE ISNOWA
%token GIMMEH OUTTADATABUKKIT
%token ORLY OIC YARLY NOWAI MEBBE
%token WTF OMG OMGWTF GTFO IMINYR IMOUTTAYR

/* Debugging tokens. These aren't part of the language, but I'm sticking 'em in to help me debug stuff */
%token YRVARZ

/* scan.l doesn't, at ths point, ever use these tokens, but we use them as types
 * of node in the parse "tree", so I'm declaring them here (instead of as separate
 * DEFINEs) so that their integer equivalents don't end up matching one of the other
 * tokens */
%token IDENT CMDBLOCK DECL MEBBEBLOCK

%%
lol_program:
	HAI EXPREND lol_cmdblock KTHXBYE EXPREND
				{ executeProgram($3); }
	;

lol_cmdblock:
	lol_cmdblock lol_cmd	{ $$ = lolListAppend($1, $2); }
	| lol_cmd 		{ $$ = lolMakeList($1); }
	;

lol_cmd:
	lol_expression EXPREND	{ $$ = $1; }
	| ORLY EXPREND lol_orly_block OIC EXPREND
				{ $$ = lolMakeNode(ORLY, tmpnval, $3); }
	| WTF EXPREND lol_wtf_block OIC EXPREND
				{ $$ = lolMakeNode(WTF, tmpnval, $3); }
	| IMINYR IDENTIFIER EXPREND lol_cmdblock IMOUTTAYR IDENTIFIER EXPREND
				{ 
					/* TODO: Check $2 and $7 for equality */
					$$ = lolMakeNode(IMINYR, tmpnval, $4);
				}
	;

lol_expression:
		/* TODO: From the spec: It has infinite arity and implicitly
		 * concatenates all of its arguments after casting them to
		 * YARNs. It is terminated by the statement delimiter (line end
		 * or comma).  The output is automatically terminated with a
		 * carriage return (:)), unless the final token is terminated
		 * with an exclamation point (!), in which case the carriage
		 * return is suppressed. */

	VISIBLE lol_visibleLevel lol_expression
				{
					tmpnval.numbrVal = $2;
					$$ = lolMakeNode(VISIBLE, tmpnval, lolMakeList($3));
				}
	| VISIBLE lol_expression
				{ 
					tmpnval.numbrVal = NOTICE;
					$$ = lolMakeNode(VISIBLE, tmpnval, lolMakeList($2));
				}
	| FOUNDYR lol_expression	{ $$ = lolMakeNode(FOUNDYR, tmpnval, lolMakeList($2)); }
	| IHASA lol_var ITZ lol_expression
				{
					$$ = lolMakeNode(DECL, ($2)->nodeVal, lolMakeList($4));
				}
	| IHASA lol_var		{ $$ = lolMakeNode(DECL, ($2)->nodeVal, NULL); }
	| lol_var R lol_expression
				{ $$ = lolMakeNode(R, ($1)->nodeVal, lolMakeList($3)); }
		/* Note that MAEK syntax is MAEK lol_expression, not MAEK lol_var, because
		   MAEK casts the value of an expression, and changes the type of IT */
	| MAEK lol_expression lol_typename
				{
					tmpnval.numbrVal = $3;
					$$ = lolMakeNode(MAEK, tmpnval, lolMakeList($2));
				}
	| MAEK lol_expression A lol_typename
				{
					tmpnval.numbrVal = $4;
					$$ = lolMakeNode(MAEK, tmpnval, lolMakeList($2));
				}
	| lol_var ISNOWA lol_typename
				{
					tmpnval.numbrVal = $3;
					$$ = lolMakeNode(ISNOWA, tmpnval, lolMakeList($1));
				}
	| GIMMEH YRVARZ		{ $$ = lolMakeNode(YRVARZ, tmpnval, NULL); }
	| GIMMEH lol_var OUTTADATABUKKIT lol_expression
				{
					$$ = lolMakeNode(OUTTADATABUKKIT, ($2)->nodeVal, lolMakeList($4));
				}
	| GIMMEH OUTTADATABUKKIT lol_expression
				{
					tmpnval.yarnVal = pstrdup("IT");
					$$ = lolMakeNode(OUTTADATABUKKIT, tmpnval, lolMakeList($3));
				}
	| GTFO			{ $$ = lolMakeNode(GTFO, tmpnval, NULL); }
	| lol_xaryoprgroup MKAY	{ $$ = $1; }
	| lol_binaryopr lol_expression AN lol_expression
				{
					$$ = lolMakeNode($1, tmpnval, lolMakeTwoNodeList($2, $4));
				}
	| lol_binaryopr lol_expression lol_expression
				{
					$$ = lolMakeNode($1, tmpnval, lolMakeTwoNodeList($2, $3));
				}
	| lol_unaryopr lol_expression	{ $$ = lolMakeNode($1, tmpnval, lolMakeList($2)); }
	| lol_xaryoprgroup 	{ $$ = $1; }
	| lol_const		{ $$ = $1; }
	| lol_var		{ $$ = $1; }
	;

lol_orly_block:
	YARLY EXPREND lol_cmdblock	{ $$ = lolMakeList(lolMakeNode(YARLY, tmpnval, $3)); }
		/* Structure of a MEBBE node in the parse tree:
		*	The MEBBE node points to a two-node list. The first node is the parse tree of the MEBBE expression.
		*	The second node is a MEBBEBLOCK node, whose list points to the lol_cmdblock to be executed if the
		*	expression is true */
	| lol_orly_block MEBBE lol_expression EXPREND lol_cmdblock
					{ $$ = lolListAppend($1, lolMakeNode(MEBBE, tmpnval, lolMakeTwoNodeList($3, lolMakeNode(MEBBEBLOCK, tmpnval, $5)))); }
	| lol_orly_block NOWAI EXPREND lol_cmdblock
					{ $$ = lolListAppend($1, lolMakeNode(NOWAI, tmpnval, $4)); }
	;

lol_wtf_block:
	OMG lol_const EXPREND lol_cmdblock
				{
					($2)->list = $4;
					tmpnval.numbrVal = 0;
					$$ = lolMakeList(lolMakeNode(OMG, tmpnval, lolMakeList($2)));
				}
				/*
	| OMG lol_const EXPREND lol_cmdblock GTFO EXPREND
				{
					($2)->list = $4;
					tmpnval.numbrVal = 1;
					$$ = lolMakeList(lolMakeNode(OMG, tmpnval, lolMakeList($2)));
				} */
	| lol_wtf_block OMG lol_const EXPREND lol_cmdblock
				{
					($3)->list = $5;
					tmpnval.numbrVal = 0;
					$$ = lolListAppend($1, lolMakeNode(OMG, tmpnval, lolMakeList($3)));
				}
				/* 
	| lol_wtf_block OMG lol_const EXPREND lol_cmdblock GTFO EXPREND
				{
					($3)->list = $5;
					tmpnval.numbrVal = 1;
					$$ = lolListAppend($1, lolMakeNode(OMG, tmpnval, lolMakeList($3)));
				}
				*/
	| lol_wtf_block OMGWTF EXPREND lol_cmdblock
				{ $$ = lolListAppend($1, lolMakeNode(OMGWTF, tmpnval, $4)); }
	;

lol_typename:
	TROOFTYPE		{ $$ = ident_TROOF; }
	| NUMBRTYPE		{ $$ = ident_NUMBR; }
	| NUMBARTYPE		{ $$ = ident_NUMBAR; }
	| YARNTYPE		{ $$ = ident_YARN; }
	| NOOBTYPE		{ $$ = ident_NOOB; }
	;

lol_var:
	IDENTIFIER		{
					tmpnval.yarnVal = pstrdup(pllolcode_yytext);
					$$ = lolMakeNode(IDENT, tmpnval, NULL);
				}
	;

	/* Most notations refer to these operators as n-ary. Calling the
	 * related functions naryOpr* makes them 
	 * too close in name to binaryOpr* and especially unaryOpr* for my
	 * liking -- too easy to confuse them that way. So I'm calling these
	 * operators x-ary operators */

lol_xaryoprgroup:
	lol_xaryopr lol_expression
				{ $$ = lolMakeNode($1, tmpnval, lolMakeList($2)); }
	| lol_xaryoprgroup lol_expression	{ lolListAppend(($1)->list, $2); $$ = $1; }
	| lol_xaryoprgroup AN lol_expression	{ lolListAppend(($1)->list, $3); $$ = $1; }
	;

lol_xaryopr:
	ALLOF		{ $$ = ALLOF; }
	| ANYOF		{ $$ = ANYOF; }
	;

lol_unaryopr:
	NOT		{ $$ = NOT; }
	;

lol_binaryopr:
	SUMOF		{ $$ = SUMOF; }
	| DIFFOF	{ $$ = DIFFOF; }
	| PRODUKTOF	{ $$ = PRODUKTOF; }
	| QUOSHUNTOF	{ $$ = QUOSHUNTOF; }
	| MODOF		{ $$ = MODOF; }
	| BIGGROF	{ $$ = BIGGROF; }
	| SMALLROF	{ $$ = SMALLROF; }
	| BOTHOF	{ $$ = BOTHOF; }
	| EITHEROF	{ $$ = EITHEROF; }
	| WONOF		{ $$ = WONOF; }
	| BOTHSAEM	{ $$ = BOTHSAEM; }
	| DIFFRINT	{ $$ = DIFFRINT; }
	;

lol_const:
	YARN			{
					tmpnval.yarnVal = pstrdup($1);
					$$ = lolMakeNode(YARN, tmpnval, NULL);
				}
	| NUMBAR		{ tmpnval.numbarVal = $1; $$ = lolMakeNode(NUMBAR, tmpnval, NULL); }
	| NUMBR			{ tmpnval.numbrVal = $1; $$ = lolMakeNode(NUMBR, tmpnval, NULL); }
	| TROOFWIN		{ tmpnval.numbrVal = 1; $$ = lolMakeNode(TROOF, tmpnval, NULL); }
	| TROOFFAIL		{ tmpnval.numbrVal = 0; $$ = lolMakeNode(TROOF, tmpnval, NULL); }
	;

lol_visibleLevel:
	VISDEBUG		{ $$ = DEBUG1; }
	| VISLOG		{ $$ = LOG; }
	| VISNOTICE		{ $$ = NOTICE; }
	| VISINFO		{ $$ = INFO; }
	| VISWARNING		{ $$ = WARNING; }
	| VISEXCEPTION		{ $$ = ERROR; }
	;
%%

void
pllolcode_yyerror(char *s)
{
    elog(ERROR, s);
}

/*extern int pllolcode_yy_flex_debug; */

bool
executeList(lolNodeList *list)
{
	lolNode *cur;
	bool gtfo = false;

	if (list == NULL) return false;
	cur = list->head;
	while (cur != NULL) {
		if (cur->node_type == GTFO) return true;
		gtfo = executeNode(cur);
		cur = cur->next;
	}
	return gtfo;
}

void
executeProgram(lolNodeList *prg)
{
	elog(DEBUG3, "Executing program");
	/*printNodeList(prg); */
	if (prg == NULL)  {
		elog(INFO, "NULL parse tree");
		return;
	}
	executeList(prg);
}

void
visibleNode(lolNode *node)
{
	/* TODO: Try printing a NOOB, and seeing what happens. Might crash the server */
	executeList(node->list);
	elog(node->nodeVal.numbrVal, lolVarGetString(lolGetVar("IT"), false));
}

void
foundyrNode(lolNode *node)
{
	executeList(node->list);
	returnVal = lolGetVar("IT");
}

void
yarnNode(lolNode *node)
{
	lolSetVar("IT", ident_YARN, node->nodeVal);
}

void
numbrNode(lolNode *node) 
{
	lolSetVar("IT", ident_NUMBR, node->nodeVal);
}

void
numbarNode(lolNode *node)
{
	lolSetVar("IT", ident_NUMBAR, node->nodeVal);
}

void
troofNode(lolNode *node)
{
	lolSetVar("IT", ident_TROOF, node->nodeVal);
}

void
identNode(lolNode *node)
{
	lolIdent *cur;
	NodeVal val;

	cur = lolGetVar(node->nodeVal.yarnVal);
	if (cur != NULL)
		lolSetVar("IT", cur->type, cur->value);
	else 
	{
		val.numbrVal = 0;
		lolSetVar("IT", ident_NUMBR, val);
	}
}

lolIdent
arithNode(lolNode *node, lolIdent a, lolIdent b)
{
	lolIdent result;
	double x, y, answer;

	result.type = ident_NUMBAR;

	/* Arithmetic with NULL values leaves non-null value unchanged */
	if (a.type == ident_NOOB) 
		return b;
	if (b.type == ident_NOOB) 
		return a;

	if (a.type == ident_NUMBR && b.type == ident_NUMBR)
		result.type = ident_NUMBR;

	x = lolVarGetNumbar(&a, false);
	y = lolVarGetNumbar(&b, false);

	switch (node->node_type) {
		case SUMOF:
			answer = x + y;
			break;
		case DIFFOF:
			answer = x - y;
			break;
		case PRODUKTOF:
			answer = x * y;
			break;
		case QUOSHUNTOF:
			if (y == 0) 
				elog(ERROR, "Division by zero");
			answer = x / y;
			break;
		case MODOF:
			if (result.type != ident_NUMBR)
				elog(DEBUG1, "Casting to integers for modulo operation");
			answer = (int) x % (int) y;
			break;
		case BIGGROF:
			answer = (x > y) ? x : y;
			break;
		case SMALLROF:
			answer = (x < y) ? x : y;
			break;
		default:
			result.type = ident_NOOB;
			answer = 0;
			elog(DEBUG1, "Strange node type passed to arithNode");
	}
	if (result.type == ident_NUMBR) 
		result.value.numbrVal = (int) answer;
	else 
		result.value.numbarVal = answer;
	return result;
}

lolIdent
boolNode(lolNode *node, lolIdent a, lolIdent b)
{
	bool x, y;
	lolIdent result;

	x = lolVarGetBool(&a);
	y = lolVarGetBool(&b);
	result.type = ident_TROOF;
	switch (node->node_type) {
		case BOTHOF:
			result.value.numbrVal = x && y;
			break;
		case EITHEROF:
			result.value.numbrVal = x || y;
			break;
		case WONOF:
			result.value.numbrVal = (x && !y) || (y && !x);
			break;
	}
	return result;
}

lolIdent
comparisonNode(lolNode *node, lolIdent a, lolIdent b)
{
	lolIdent result;
	double x, y;

	result.type = ident_TROOF;
	result.value.numbrVal = 0;
	if (a.type != b.type) {
		if ((a.type == ident_NUMBR && b.type == ident_NUMBAR) || 
			(a.type == ident_NUMBAR && b.type == ident_NUMBR)) {
			if (a.type == ident_NUMBR)
				x = NUMBR2NUMBAR(a.value.numbrVal);
			else x = a.value.numbarVal;
			if (b.type == ident_NUMBR)
				y = NUMBR2NUMBAR(b.value.numbrVal);
			else y = b.value.numbarVal;
			result.value.numbrVal = ( x == y ) ? 1 : 0;
		}
	}
	else switch (a.type) {
		case ident_TROOF:
			if ((a.value.numbrVal > 0 && b.value.numbrVal > 0) || (a.value.numbrVal == 0 && b.value.numbrVal == 0))
				result.value.numbrVal = 1;
			break;
		case ident_NUMBR:
			if (a.value.numbrVal == b.value.numbrVal)
				result.value.numbrVal = 1;
			break;
		case ident_NOOB:
			break;
		case ident_NUMBAR:
			if (a.value.numbarVal == b.value.numbarVal)
				result.value.numbrVal = 1;
			break;
		case ident_YARN:
			/* TODO: Once strings can be longer than 255 chars, change this */
			result.value.numbrVal = (strncmp(b.value.yarnVal, a.value.yarnVal, 255) == 0 ? 1 : 0);
			break;
	}
	if (node->node_type == DIFFRINT)
		result.value.numbrVal = (result.value.numbrVal == 1) ? 0 : 1;
	return result;
}

void
assignNode(lolNode *node)
{
	/* Variable name is in node->nodeVal.yarnVal, and value is in the list */
	lolIdent *IT;

	executeNode(node->list->head);
	IT = lolGetVar("IT");
	lolSetVar(node->nodeVal.yarnVal, IT->type, IT->value);
}

void
binaryOprNode(lolNode *node)
{
	lolNode *cur;
	lolIdent a, b;
	lolIdent *pIT;

	/* Assume we've got a list of two nodes. Execute each node in turn, and
	   retrieve the results in a and b */
	if (node->list == NULL) return;
	cur = node->list->head;
	executeNode(cur);
	pIT = lolGetVar("IT");
	a.type = pIT->type;
	a.value = pIT->value;
	cur = cur->next;
	if (cur == NULL) return;
	executeNode(cur);
	pIT = lolGetVar("IT");
	b.type = pIT->type;
	b.value = pIT->value;

	switch (node->node_type) {
		case SUMOF:
		case DIFFOF:
		case PRODUKTOF:
		case QUOSHUNTOF:
		case MODOF:
		case BIGGROF:
		case SMALLROF:
			a = arithNode(node, a, b);
			break;
		case BOTHOF:
		case EITHEROF:
		case WONOF:
			a = boolNode(node, a, b);
			break;
		case BOTHSAEM:
		case DIFFRINT:
			a = comparisonNode(node, a, b);
			break;
		default:
			return;
	}
	lolSetVar("IT", a.type, a.value);
}

void
unaryOprNode(lolNode *node)
{
	/* So far, NOT is the only unary operator I know of. So I'm making this very NOT-specific */
	NodeVal tmpVal;

	executeList(node->list);
	tmpVal.numbrVal = lolVarGetBool(lolGetVar("IT")) ? 0 : 1;
	lolSetVar("IT", ident_TROOF, tmpVal);
}

void
allOfNode(lolNode *node)
{
	lolNode *cur;
	NodeVal tmpVal;

	if (node->list == NULL) return;
	cur = node->list->head;
	while (cur != NULL) {
		executeNode(cur);
		if (!lolVarGetBool(lolGetVar("IT"))) {
			tmpVal.numbrVal = 0;
			lolSetVar("IT", ident_TROOF, tmpVal);
			return;
		}
		cur = cur->next;
	}
	tmpVal.numbrVal = 1;
	lolSetVar("IT", ident_TROOF, tmpVal);
}

void
anyOfNode(lolNode *node)
{
	lolNode *cur;
	NodeVal tmpVal;

	if (node->list == NULL) return;
	cur = node->list->head;
	while (cur != NULL) {
		executeNode(cur);
		if (lolVarGetBool(lolGetVar("IT"))) {
			tmpVal.numbrVal = 1;
			lolSetVar("IT", ident_TROOF, tmpVal);
			return;
		}
		cur = cur->next;
	}
	tmpVal.numbrVal = 0;
	lolSetVar("IT", ident_TROOF, tmpVal);
}

void
castNode(lolNode *node)
{
	lolIdent *var;
	NodeVal nodeVal;
	char tmpString[255];
	NodeType newType;

	/* The list one node, containing a variable name; this node's
	 * nodeVal.numbrVal is the type the variable is supposed to be cast to.
	 * The node can be a MAEK node, in which case the result gets stuck in
	 * IT, or a ISNOWA node, which means the variable itself will be
	 * modified.  Note that an ISNOWA node will also modify IT. */

	if (node->list == NULL || node->list->head == NULL) {
		elog(DEBUG1, "PL/LOLCODE: Empty node list in cast operation");
		return;
	}
	var = lolGetVar(node->list->head->nodeVal.yarnVal);
	newType = node->nodeVal.numbrVal;
	if (var->type == newType) {
		elog(DEBUG3, "PL/LOLCODE: Casting to same type");
		return;
	}

	switch (newType) {
		case ident_YARN:
			switch (var->type) {
				case ident_NUMBR:
					NUMBR2YARN(var->value.numbrVal, tmpString, 255);
					nodeVal.yarnVal = pstrdup(tmpString);
					break;
				case ident_NUMBAR:
					NUMBAR2YARN(var->value.numbarVal, tmpString, 255);	
					nodeVal.yarnVal = pstrdup(tmpString);
					break;
				case ident_TROOF:
					if (nodeVal.numbrVal)
						nodeVal.yarnVal = pstrdup("WIN");
					else
						nodeVal.yarnVal = pstrdup("FAIL");
					break;
				case ident_NOOB:
					nodeVal.yarnVal = pstrdup("");
					break;
				default:
					break;
			}
			break;
		case ident_NUMBR:
			switch (var->type) {
				case ident_NUMBAR:
					nodeVal.numbrVal = NUMBAR2NUMBR(var->value.numbarVal);
					break;
				case ident_YARN:
					nodeVal.numbrVal = YARN2NUMBR(var->value.yarnVal);
					break;
				case ident_TROOF:
					nodeVal.numbrVal = var->value.numbrVal;
					break;
				case ident_NOOB:
					nodeVal.numbrVal = 0;
					break;
				default:
					break;
			}
			break;
		case ident_NUMBAR:
			switch (var->type) {
				case ident_NUMBR:
					nodeVal.numbarVal = NUMBR2NUMBAR(var->value.numbrVal);
					break;
				case ident_YARN:
					nodeVal.numbarVal = YARN2NUMBAR(var->value.yarnVal);
					break;
				case ident_TROOF:
					nodeVal.numbarVal = (double) var->value.numbrVal;
					break;
				case ident_NOOB:
					nodeVal.numbarVal = 0;
					break;
				default:
					break;
			}
			break;
		case ident_TROOF:
			switch (var->type) {
				case ident_NUMBR:
					nodeVal.numbrVal = (var->value.numbrVal != 0) ? 1 : 0;
					break;
				case ident_NUMBAR:
					nodeVal.numbrVal = (var->value.numbarVal != 0) ? 1 : 0;
					break;
				case ident_YARN:
					nodeVal.numbrVal = (strlen(var->value.yarnVal) > 0) ? 1 : 0;
					break;
				case ident_NOOB:
					nodeVal.numbrVal = 0;
					break;
				default:
					break;
			}
			break;
		case ident_NOOB:
			nodeVal.numbrVal = 0;
			break;
		default:
			break;
	}
	lolSetVar("IT", newType, nodeVal);

	if (node->node_type == ISNOWA) 
	{
		var->type = newType;
		var->value = nodeVal;
	}
}

bool
orlyNode(lolNode *node)
{
	lolNode *cur;
	bool blockDone = 0;
	bool gtfo = false;

	if (node->list == NULL) return false;
	cur = node->list->head;

	while (cur != NULL) {
		switch (cur->node_type) {
			case YARLY:
				if (lolVarGetTroof(lolGetVar("IT"))) {
					gtfo = executeList(cur->list);
					blockDone = 1;
				}
				break;
			case MEBBE:
				if (cur->list == NULL || cur->list->head->next == NULL) {
					elog(ERROR, "Invalid construction of MEBBE node -- missing execute list for MEBBE or MEBBEBLOCK");
					return false;
				}
				executeNode(cur->list->head);
				if (lolVarGetTroof(lolGetVar("IT"))) {
					gtfo = executeNode(cur->list->head->next);
					blockDone = 1;
				}
				break;
			case NOWAI:
				if (!blockDone) {
					gtfo = executeList(cur->list);
				}
				break;
			default:
				elog(ERROR, "Invalid block type %d in O RLY parse tree", cur->node_type);
		}
		cur = cur->next;
	}
	return gtfo;
}

void
wtfNode(lolNode *node)
{
	lolNode *cur;
	bool foundMatch = 0, gtfo = 0;
	lolIdent *tmp = lolGetVar("IT");
	lolIdent myIT, compResult;
	lolNode myNode;

	if (node->list == NULL) return;
	cur = node->list->head;
	myIT.type = tmp->type;
	myIT.value = tmp->value;
	myNode.node_type = BOTHSAEM;

	while (cur != NULL && !gtfo) {
		if (cur->node_type == OMG) {
			if (cur->list == NULL) return;

			/* The code block should be executed if one of the following is true:
				1) the included literal matches the original value of IT
				2) some earlier literal matched IT and no GTFO has been reached since then */
			if (!foundMatch) { 
				/* Note: This handling will have to be reworked when I make IT behave properly. I can't just go resetting IT all over */
				executeNode(cur->list->head);
				tmp = lolGetVar("IT");
				compResult = comparisonNode(&myNode, myIT, *tmp);
			}

			if ((foundMatch && !gtfo) || compResult.value.numbrVal == 1) {
				foundMatch = 1;
				lolSetVar("IT", myIT.type, myIT.value);
				gtfo = executeList(cur->list->head->list);
			}
		}
		else if (cur->node_type == OMGWTF) {
			if (!foundMatch || (foundMatch && !gtfo)) {
				executeList(cur->list);
			}
		}
		else {
			elog(ERROR, "Unknown node type %d in WTF list in parse tree", node->node_type);
			return;
		}
		cur = cur->next;
	}
}

void
iminyrNode(lolNode *node)
{
	bool endLoop = false;

	if (node == NULL) return;

	while (!endLoop)
		endLoop = executeList(node->list);
}

bool
executeNode(lolNode *node)
{
	bool gtfo = false;
	lolIdent *tmpVar;
	if (node == NULL) return false;
	switch (node->node_type) {
		case VISIBLE:
			visibleNode(node);
			break;
		case YARN:
			yarnNode(node);
			break;
		case NUMBR:
			numbrNode(node);
			break;
		case NUMBAR:
			numbarNode(node);
			break;
		case TROOF:
			troofNode(node);
			break;
		case FOUNDYR:
			foundyrNode(node);
			break;
		case IDENT:
			identNode(node);
			break;	
		case SUMOF:
		case DIFFOF:
		case PRODUKTOF:
		case QUOSHUNTOF:
		case MODOF:
		case BIGGROF:
		case SMALLROF:
		case BOTHOF:
		case EITHEROF:
		case WONOF:
		case BOTHSAEM:
		case DIFFRINT:
			binaryOprNode(node);
			break;
		case NOT:
			unaryOprNode(node);
			break;
		case ALLOF:
			allOfNode(node);
			break;
		case ANYOF:
			anyOfNode(node);
			break;
		case DECL:
			/* Variable name is in this nodes value, node->list, if not null, contains the expression to assign it to */
			lolDeclareVar(node->nodeVal.yarnVal);
			if (node->list != NULL) {
				executeNode(node->list->head);
				tmpVar = lolGetVar("IT");
				lolSetVar(node->nodeVal.yarnVal, tmpVar->type, tmpVar->value);
			}			
			break;
		case MAEK:
		case ISNOWA:
			castNode(node);
			break;
		case R:
			assignNode(node);
			break;
		case OUTTADATABUKKIT:
			SPInode(node);
			break;
		case ORLY:
			gtfo = orlyNode(node);
			break;
		case MEBBEBLOCK:
			executeList(node->list);
			break;
		case WTF:
			wtfNode(node);
			break;
		case IMINYR:
			iminyrNode(node);
			break;

			/* Debugging stuff */
		case YRVARZ:
			printSymbolTable();
			break;

		default:
			elog(INFO, "Cannot execute node type");
			printNode(node);
			break;
	}
	return gtfo;
}

lolNodeList *
lolMakeTwoNodeList(lolNode *node1, lolNode *node2)
{
	lolNodeList *list = lolMakeList(node1);
	return lolListAppend(list, node2);
}

lolNodeList *
lolListAppend(lolNodeList *list, lolNode *node)
{
	if (list == NULL) { elog(DEBUG1, "list is null in lolListAppend"); return NULL; }
	if (node == NULL) { elog(DEBUG1, "node is null in lolListAppend"); return NULL; }
	node->next = NULL;
	if (list->tail == NULL) {
		list->head = node;
		list->tail = node;
		list->length = 1;
		return list;
	}
	list->tail->next = node;
	list->tail = node;
	list->length++;
	if (debugList) printNodeList(list);
	return list;
}

lolNodeList *
lolMakeList(lolNode *node)
{
	lolNodeList *list;
	
	list = palloc(sizeof(lolNodeList));
	list->head = node;
	list->tail = node;
	list->length = 1;
	if (debugList) printNodeList(list);
	return list;	
} 
lolNode *
lolMakeNode(NodeType typ, NodeVal nodeVal, lolNodeList *list)
{
	lolNode *node;

	node = palloc(sizeof(lolNode));
	node->node_type = typ;
	node->nodeVal = nodeVal;
	node->list = list;
	node->next = NULL;
	return node;
}
void
printNode(lolNode *cur)
{
	char tmpString[255];

	if (cur == NULL) {
		elog(INFO, "NULL node");
		return;
	}
	switch (cur->node_type) {
		case CMDBLOCK:
			elog(INFO, "CMDBLOCK");
			printNodeList(cur->list);
			break;
		case VISIBLE:
			elog(INFO, "VISIBLE");
			printNodeList(cur->list);
			break;
		case YARN:
			elog(INFO, "YARN");
			snprintf(tmpString, 255, "value: %s", cur->nodeVal.yarnVal);
			elog(INFO, tmpString);
			break;
		case NUMBR:
			elog(INFO, "NUMBR");
			snprintf(tmpString, 255, "value: %d", cur->nodeVal.numbrVal);
			elog(INFO, tmpString);
			break;
		case NUMBAR:
			elog(INFO, "NUMBAR");
			snprintf(tmpString, 255, "value: %f", cur->nodeVal.numbarVal);
			elog(INFO, tmpString);
			break;
		case FOUNDYR:
			elog(INFO, "FOUNDYR");
			printNodeList(cur->list);
			break;
		case IDENT:
			elog(INFO, "IDENT: name = %s", cur->nodeVal.yarnVal);
			break;	
		case ORLY:
			elog(INFO, "ORLY");
			printNodeList(cur->list);
			break;
		case YARLY:
			elog(INFO, "YA RLY");
			printNodeList(cur->list);
			break;
		case MEBBE:
			elog(INFO, "MEBBE");
			printNodeList(cur->list);
			break;
		case MEBBEBLOCK:
			elog(INFO, "MEBBEBLOCK");
			printNodeList(cur->list);
			break;
		case NOWAI:
			elog(INFO, "NOWAI");
			printNodeList(cur->list);
			break;
		case BOTHSAEM:
			elog(INFO, "BOTH SAEM");
			break;
		case WTF:
			elog(INFO, "WTF");
			printNodeList(cur->list);
			break;
		case OMG:
			elog(INFO, "OMG");
			if (cur->nodeVal.numbrVal) elog(INFO, "  This OMG node has a GTFO");
			printNode(cur->list->head);
			if (cur->list != NULL && cur->list->head != NULL)
				printNodeList(cur->list->head->list);
			else elog(INFO, "ERROR: Badly constructed OMG node");
			break;
		case OMGWTF:
			elog(INFO, "OMGWTF");
			printNodeList(cur->list);
			break;
		default:
			elog(INFO, "Unknown node type %d", cur->node_type);
	}
}

void
printNodeList(lolNodeList *list)
{
	lolNode *cur;

	elog(INFO, "Start of list");
	if (list == NULL) 
	{
		elog(INFO, "NULL node list");
		return;
	}
	cur = list->head;
	if (cur == NULL) {
		elog(INFO, "NULL first node");
		return;
	}
	while (cur->next != NULL) {
		elog(INFO, "List Node");
		printNode(cur);
		cur = cur->next;
	}
	printNode(cur);
	elog(INFO, "End of list");
}

void
SPInode(lolNode *node) 
{
	char *SPIval;

	/*| GIMMEH lol_var OUTTADATABUKKIT lol_expression
				{
					$$ = lolMakeNode(OUTTADATABUKKIT, ($2)->nodeVal, lolMakeList($4)); */
	int res;
	if (node->list == NULL) {
		elog(NOTICE, "GIMMEH ... OUTTA DATABUKKIT needs a query to pass to the databukkit.");
		return;
	}
	executeList(node->list);
	/* The query is now in IT */
	res = SPI_exec(lolVarGetString(IT, false), 0);
	switch (res) {
		case SPI_OK_SELECT:
		case SPI_OK_INSERT_RETURNING:
		case SPI_OK_DELETE_RETURNING:
		case SPI_OK_UPDATE_RETURNING:
			elog(DEBUG3, "PL/LOLCODE SPI: Processed %d rows", SPI_processed);
			if (SPI_processed < 1) {
				elog(DEBUG5, "PL/LOLCODE SPI: No rows returned");
				return;
			}
			SPIval = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1);
			LOLifyString(SPIval, SPI_gettypeid(SPI_tuptable->tupdesc, 1), node->nodeVal.yarnVal);
			break;

		case SPI_ERROR_ARGUMENT:
			elog(ERROR, "SPI_execute returned SPI_ERROR_ARGUMENT. Please provide a proper query to retrieve.");
			break;
		case SPI_ERROR_COPY:
			elog(ERROR, "PL/LOLCODE can't copy to STDOUT or from STDIN");
			break;
		case SPI_ERROR_TRANSACTION:
			elog(ERROR, "Transaction's cannot be manipulated using GIMMEH ... OUTTA DATABUKKIT");
			break;
		case SPI_ERROR_OPUNKNOWN:
			elog(ERROR, "Something bad's going on -- SPI return SPI_ERROR_OPUNKNOWN.");
			break;
		case SPI_ERROR_UNCONNECTED:
			elog(ERROR, "Disconnected from SPI. No idea why.");
			break;
		default:
			break;
	}
}
